#define LON_TIMEOUT	300
#define MAX_INTEL_HEX_RECORD_LENGTH 16
typedef struct _INTEL_HEX_RECORD {
	__u16	length;
	__u16	address;
	__u8	type;
	__u8	data[MAX_INTEL_HEX_RECORD_LENGTH];
} INTEL_HEX_RECORD, *PINTEL_HEX_RECORD;

#define FIRMWARE_LOADER	"gesytec/loader.fw"
#define FIRMWARE_FX 	"gesytec/lonusb_fx.fw"
#define FIRMWARE_FX2	"gesytec/lonusb_fx2.fw"

MODULE_FIRMWARE(FIRMWARE_LOADER);
MODULE_FIRMWARE(FIRMWARE_FX);
MODULE_FIRMWARE(FIRMWARE_FX2);

#define ANCHOR_LOAD_INTERNAL	0xA0	/* Vendor specific request code for Anchor Upload/Download (This one is implemented in the core) */
#define ANCHOR_LOAD_EXTERNAL	0xA3	/* This command is not implemented in the core. Requires firmware */
#define ANCHOR_LOAD_FPGA	    0xA5	/* This command is not implemented in the core. Requires firmware. Emagic extension */
#define MAX_INTERNAL_ADDRESS	0x1B3F	/* This is the highest internal RAM address for the AN2131Q */
#define CPUCS_REG		        0x7F92	/* EZ-USB Control and Status Register.  Bit 0 controls 8051 reset */
#define CPUCS_REG_FX2           0xE600
#define INTERNAL_RAM(address)   (address <= MAX_INTERNAL_ADDRESS)


// thanks to drivers/usb/serial/keyspan_pda.c code
static int ezusb_writememory(struct usb_device *dev, int addr, const unsigned char *data, int len, __u8 request)
{
	int result;
	void *buf = kmemdup(data, len, GFP_KERNEL);
	if (!buf)
		return -ENOMEM;

	result = usb_control_msg(dev, usb_sndctrlpipe(dev, 0), request, 0x40, addr, 0, buf, len, LON_TIMEOUT);
	kfree(buf);
	return result < 0 ? result : 0;
}

static int ezusb_set_reset(struct usb_device *dev, unsigned char reset_bit)
{
	// toggle the EZ-USB reset bit (harmless on FX2)
	int ret = ezusb_writememory(dev, CPUCS_REG, &reset_bit, 1, 0xa0);
	if (ret < 0) 
    {
		err("ezusb: set_reset (%d) failed", reset_bit);
//		return ret;
	}
	// toggle the FX2 reset bit (harmless on EZ-USB)
	ret = ezusb_writememory(dev, CPUCS_REG_FX2, &reset_bit, 1, 0xa0);
	if (ret < 0)
		err("ezusb: set_reset (%d) failed", reset_bit);
	return ret;
}

static int ezusb_load_firmware(struct usb_device *dev, const void *fw)
{
	const INTEL_HEX_RECORD *hexRecord = fw;
	const struct firmware *loader_fw;
	const INTEL_HEX_RECORD *loader_hx;
	int i;

	int err = request_firmware(&loader_fw, FIRMWARE_LOADER, &dev->dev);
	if (err < 0) 
    {
		err("Failed to load %s: %d\n", FIRMWARE_LOADER, err);
		return err;
	}

	// Assert reset (stop the CPU in the ezusb)
	err = ezusb_set_reset(dev, 1);
	if (err < 0) 
    {
        err("%s - error(1) loading firmware: error = %d", __FUNCTION__, err);
		goto err_fw;
	}
	loader_hx = (const INTEL_HEX_RECORD *)loader_fw->data;
	// 1. We need to put the loader for the firmware into the EZ-USB
	for (i = 0; loader_hx[i].type == 0; i++) 
    {
		err = ezusb_writememory(dev, le16_to_cpu(loader_hx[i].address), loader_hx[i].data, le16_to_cpu(loader_hx[i].length), ANCHOR_LOAD_INTERNAL);
		if (err < 0) 
        {
            err("%s - error(2) loading firmware: error = %d", __FUNCTION__, err);
			goto err_fw;
		}
	}

	// De-assert reset (let the CPU run)
	err = ezusb_set_reset(dev, 0);
	if (err < 0) 
    {
        err("%s - error(3) loading firmware: error = %d", __FUNCTION__, err);
		goto err_fw;
	}

	// 2. We put the part of the firmware that lies in the external RAM into the EZ-USB
	for (i = 0; hexRecord[i].type == 0; i++) 
    {
        if (!INTERNAL_RAM(le16_to_cpu(hexRecord[i].address)))
        {
            err = ezusb_writememory(dev, le16_to_cpu(hexRecord[i].address), hexRecord[i].data, le16_to_cpu(hexRecord[i].length), ANCHOR_LOAD_EXTERNAL);
			if (err < 0) 
            {
                err("%s - error(4) loading firmware: error = %d", __FUNCTION__, err);
				goto err_fw;
			}
		}
	}
	// Assert reset (stop the CPU in the ezusb)
	err = ezusb_set_reset(dev, 1);
	if (err < 0) 
    {
        err("%s - error(5) loading firmware: error = %d", __FUNCTION__, err);
		goto err_fw;
	}
	for (i = 0; hexRecord[i].type == 0; i++) 
    {
        if (INTERNAL_RAM(le16_to_cpu(hexRecord[i].address)))
        {
            err = ezusb_writememory(dev, le16_to_cpu(hexRecord[i].address), hexRecord[i].data, le16_to_cpu(hexRecord[i].length), ANCHOR_LOAD_INTERNAL);
			if (err < 0) 
            {
                err("%s i=%d - error(6) loading firmware: error = %d", __FUNCTION__, i, err);
				goto err_fw;
			}
		}
	}

	err = ezusb_set_reset(dev, 0);
	if (err < 0) 
    {
		err("%s - error(7) loading firmware: error = %d", __FUNCTION__, err);
		goto err_fw;
	}

	err = 0;
err_fw:
	release_firmware(loader_fw);
	return err;
}
