/*****************************************************************************/
/*
 * Linux driver for Easylon ISA/PC-104/PCI Interfaces
 *
 * Copyright (c)2004 Gesytec GmbH, Volker Schober (info@gesytec.de)
 * Licensend for use with Easylon USB Interfaces by Gesytec
 *
 * Please compile lpclinux.c
 *
 */
 /*****************************************************************************/

#include <linux/module.h>
#include <linux/init.h>
#include <linux/version.h>
#include <linux/kernel.h>
#include <linux/slab.h>

#include <linux/fs.h>
#include <linux/major.h>

#include <linux/workqueue.h>

#include <linux/ioport.h>
#include <linux/delay.h>
#include <linux/proc_fs.h>
#include <linux/timer.h>
#include <linux/poll.h>
#include <linux/pci.h>
#include <linux/usb.h>

#include "GesytecIDs.h"
#include "Plx9050.h"

/* Version Information */
#define DRIVER_VERSION  "V3.00"
#define DRIVER_AUTHOR   "Volker Schober, vschober@gesytec.de"
#define DRIVER_DESC     "Easylon LPC/LPP PnP Driver"
#define DRIVER_MOD_NAME "lpclinux"  /* module name */
#define DRIVER_DEV_NAME "lpc"
#define CLASS_NAME      "easylon"


#define IOCTL_VERSION   0x43504C00
#define IOCTL_WATCHER   0x43504C01

//#define PFX			DRIVER_MOD_NAME ": "

#define NUM_INBUF   64
#define NUM_OUTBUF  16
#define MAX_LPCS    32
#define MAX_LPPS    32

#define MAX_DEVICES (MAX_LPCS + MAX_LPPS)


struct lpc_dev {
	wait_queue_head_t rd_wait;
	spinlock_t lock;
	struct mutex mutex;
	uint	minor;
	int	open_count;
	bool	is_isa;
	struct timer_list drvtimer;

	u32 PlxIntCsrPort;
	u32 cport;
	u32 rport;
	u32 wport;
	u32 wtcport;
	unsigned int irq;

	int inhead;
	int intail;
	int outhead;
	int outtail;
	u8 ctl;
	u8 state;
	u8 active;
	struct device *device;
	u8  wtc_exist;
	u8  inbuf[NUM_INBUF][256];
	u8  outbuf[NUM_OUTBUF][256];
};

#define CMD_NULL		0x00
#define CMD_XFER		0x01
#define CMD_RESYNC		0x5A
#define CMD_ACKSYNC		0x07
#define	EOM				0x00

enum {
	ST_RESYNC = 0,
	ST_RESET,
	ST_FLCAN,
	ST_UPLINK,
	ST_IDLE,
	ST_NIACK,
};
