#define RetErrnoName(x)  case x: return #x

static inline char *ErrnoName(int ch)
    {
    static char RetKeyName[3] = {0};
    switch (ch)
        {
//        case EOF: return "EOF";
        RetErrnoName(ENOENT);
        RetErrnoName(ECONNRESET);
        RetErrnoName(EILSEQ);
        RetErrnoName(ESHUTDOWN);
        default: return RetKeyName;
        }
    }


/**
 *      lonusb_
 */
static unsigned lonusb_poll(struct file *file, struct poll_table_struct *wait)
{
	PDEVICE_EXTENSION pdx = file->private_data;
	unsigned mask = 0;

    dbg("lonusb_poll()");
	pdx->DevStats.poll++;

    /* verify that the device wasn't unplugged */
	if (pdx->usbdev == NULL || pdx->remove_pending)
	{
		pdx->DevStats.poll_error++;
	    dbgz(DEBUG_OPEN_IOCTL, "lonusb_poll(1)  return POLLERR | POLLHUP");
		return POLLERR | POLLHUP;
	}

	if (file->f_mode & FMODE_READ)
	{
		pdx->DevStats.poll_read++;
        poll_wait(file, &pdx->rd_wait, wait);
    }
	if (file->f_mode & FMODE_WRITE)
	{
		pdx->DevStats.poll_write++;
        poll_wait(file, &pdx->wr_wait, wait);
    }

	/* verify that the device wasn't unplugged */
	if (pdx->usbdev == NULL || pdx->remove_pending)
	{
		pdx->DevStats.poll_error++;
	    dbgz(DEBUG_OPEN_IOCTL, "lonusb_poll(2)  return POLLERR | POLLHUP");
		return POLLERR | POLLHUP;
	}

    if (qhasdata(&pdx->ReadBufferQueue))
	{
		pdx->DevStats.poll_pollin++;
		mask |= POLLIN | POLLRDNORM;
	}
#if REUSE_URB
    if (pdx->write_urb->status != -EINPROGRESS)
#else
	if (!atomic_read(&pdx->write_count))
#endif
	{
		pdx->DevStats.poll_pollout++;
		mask |= POLLOUT | POLLWRNORM;
	}

    dbg("lonusb_poll(3)  return %04X", mask);
	return mask;
}

static void lonusb_read_int_callback(struct urb *urb)
{
	PDEVICE_EXTENSION pdx = urb->context;
    NTSTATUS status;
    unsigned int TransferLen;
	int ret;

    switch (urb->status)
    {
    case 0:
        /* success */
        break;
    case -ECONNRESET:
    case -ENOENT:
    case -ESHUTDOWN:
        /* this urb is terminated, clean up */
        pdx->RemainDataSize = 0;
        return;
    default:
        pdx->RemainDataSize = 0;
        goto exit;
    }

	TransferLen = urb->actual_length;
	TransferLen += pdx->RemainDataSize;

	if (TransferLen == 0)
		goto exit;

	if (TransferLen > sizeof(pdx->intdata))
    {
		dbg("TransferLen=%u > sizeof(dev->intdata)=%zu", TransferLen, sizeof(pdx->intdata));
		TransferLen = pdx->RemainDataSize = 0;
		goto exit;
	}
	memcpy(&pdx->intdata[pdx->RemainDataSize], pdx->intr_in_buff, urb->actual_length);
	if (urb->actual_length == pdx->bulk_in_size)
    {
		pdx->RemainDataSize = TransferLen;
		dbg(" RemainDataSize=%u:", pdx->RemainDataSize);
		goto exit;
	}
	pdx->RemainDataSize = 0;

    DumpBuffer(DEBUG_DUMP2, "<< ", TransferLen, pdx->intdata);
    pdx->DevStats.usb_in++;
    status = qwrite(&pdx->ReadBufferQueue, pdx->intdata, TransferLen, FALSE);
    wake_up_interruptible(&pdx->rd_wait);
    dbg("wake_up_interruptible");
exit:
	ret = usb_submit_urb(urb, GFP_ATOMIC);
	if (ret)
        err("%s(): usb_submit_urb()=%d", __FUNCTION__, ret);
}


static ssize_t lonusb_read(struct file *file, char __user *buffer, size_t count, loff_t *ppos)
{
	PDEVICE_EXTENSION pdx = file->private_data;
	int ret = 0;

    dbgz(DEBUG_DISPATCH_READ, "%s(): count=%zu", __FUNCTION__, count);
	pdx->DevStats.read_in++;

	mutex_lock(&pdx->mutex);
	if (pdx->usbdev == NULL || pdx->remove_pending)
    {
		mutex_unlock(&pdx->mutex);
        pdx->DevStats.read_error++;
        dbgz(DEBUG_OPEN_IOCTL, "%s()=-ENODEV !!!", __FUNCTION__);
		return -ENODEV;
	}

    if (qempty(&pdx->ReadBufferQueue))
  {
		mutex_unlock(&pdx->mutex);
		if (file->f_flags & O_NONBLOCK)
        {
			dbgz(DEBUG_OPEN_IOCTL, "%s()=-EAGAIN !!!", __FUNCTION__);
            pdx->DevStats.read_error++;
			return -EAGAIN;
        }

        ret = wait_event_interruptible(pdx->rd_wait, (qhasdata(&pdx->ReadBufferQueue) || pdx->remove_pending));
        dbgz(DEBUG_DISPATCH_READ, "lonusb_read(): wait_event_interruptible()=%d", ret);
		if (ret < 0)
        {
            pdx->DevStats.read_error++;
			return ret;
        }
		mutex_lock(&pdx->mutex);
		if (pdx->usbdev == NULL || pdx->remove_pending)
        {
			mutex_unlock(&pdx->mutex);
            pdx->DevStats.read_error++;
            dbgz(DEBUG_OPEN_IOCTL, "%s()=-ENODEV !!!", __FUNCTION__);
			return -ENODEV;
		}
	}

    // if the read was successful, copy the data to userspace
    if (qhasdata(&pdx->ReadBufferQueue))
    {
        ret = qread(&pdx->ReadBufferQueue, buffer, count, TRUE);
        DumpBuffer(DEBUG_DUMP1, "<  ", ret, buffer);
    }

	mutex_unlock(&pdx->mutex);
    dbgz(DEBUG_DISPATCH_READ, "%s() return %d", __FUNCTION__, ret);
    pdx->DevStats.read_ok++;
	return ret;
}

static void lonusb_write_bulk_callback(struct urb *write_urb)
{
	PDEVICE_EXTENSION pdx = write_urb->context;

	DumpBuffer(DEBUG_DUMP2, ">> ", write_urb->transfer_buffer_length, write_urb->transfer_buffer);
#if !REUSE_URB
	atomic_dec(&pdx->write_count);
#endif
	wake_up_interruptible(&pdx->wr_wait);

	if (write_urb->status != 0)  /* USB_ST_NOERROR */
    {
        pdx->DevStats.write_error++;
        err("%s() write bulk error: %d=%s", __FUNCTION__, write_urb->status, ErrnoName(-write_urb->status));
        return;
    }
    else
        pdx->DevStats.write_ok++;
    dbg("%s() write bulk done", __FUNCTION__);
}

static ssize_t lonusb_write(struct file *file, const char __user *buffer, size_t count, loff_t *ppos)
{
	PDEVICE_EXTENSION pdx = file->private_data;
	struct usb_device *usbdev;
	struct urb *write_urb = NULL;
	ssize_t bytes_written = 0;
	int retval = 0;
	int offset = 0;

    pdx->DevStats.write_in++;
	mutex_lock(&pdx->mutex);
	usbdev = pdx->usbdev;
	if (usbdev == NULL || pdx->remove_pending)
    {
        dbgz(DEBUG_OPEN_IOCTL, "%s() dev->usbdev=%p dev->remove_pending=%X", __FUNCTION__, usbdev, pdx->remove_pending);
		retval = -ENODEV;
		goto error_up;
	}

	if (count == 0)
    {
        dbgz(DEBUG_DISPATCH_WRITE, "%s() write request of 0 bytes", __FUNCTION__);
		goto error_up;
    }

	offset = 0;
	while (offset < count)
    {
		/* see if we are already in the middle of a write */
#if REUSE_URB
        if (pdx->write_urb->status == -EINPROGRESS)
#else
		if (atomic_read(&pdx->write_count))
#endif
        {
			dbgz(DEBUG_DISPATCH_WRITE, "%s(): write_urb->status == -EINPROGRESS", __FUNCTION__);
			mutex_unlock(&pdx->mutex);
			if (file->f_flags & O_NONBLOCK)
			{
				dbgz(DEBUG_DISPATCH_WRITE, "%s()=-EAGAIN !!!", __FUNCTION__);
				return -EAGAIN;
			}

#if REUSE_URB
            if (wait_event_interruptible(pdx->wr_wait, pdx->write_urb->status != -EINPROGRESS))
#else
			if (wait_event_interruptible(pdx->wr_wait, !atomic_read(&pdx->write_count)))
#endif
            {
				return -ERESTARTSYS;
			}
			mutex_lock(&pdx->mutex);
		}
#if REUSE_URB
        write_urb = pdx->write_urb;
#else
		atomic_inc(&pdx->write_count);
		write_urb = usb_alloc_urb(0, GFP_KERNEL);
		if (!write_urb)
        {
			err("No free (write) urbs available");
			retval = -ENOMEM;
			goto error_up;
		}
#endif
		bytes_written = ((count-offset) > pdx->bulk_out_size) ? pdx->bulk_out_size : count-offset;

		/* copy the data (from userspace into our urb */
		if (copy_from_user(pdx->bulk_out_buffer, buffer + offset, bytes_written))
        {
			retval = -EFAULT;
			goto error;
		}
		offset += bytes_written;
		DumpBuffer(DEBUG_DUMP1, ">  ", bytes_written, pdx->bulk_out_buffer);

		usb_fill_bulk_urb(write_urb, usbdev,
			usb_sndbulkpipe(usbdev, pdx->bulk_out_endpointAddr),
			pdx->bulk_out_buffer, bytes_written,
			lonusb_write_bulk_callback, pdx);

#if REUSE_URB
        write_urb->status = 0;	// needed!
#endif
		retval = usb_submit_urb(write_urb, GFP_ATOMIC);
		if (retval)
        {
			err("failed submitting write urb, %d", retval);
		} else
        {
			retval = offset;
			goto exit;
		}
	}
error:
#if REUSE_URB
exit:
#else
	atomic_dec(&pdx->write_count);
exit:
	usb_free_urb(write_urb);
#endif
error_up:
	mutex_unlock(&pdx->mutex);
	return retval;
}

static long lonusb_unlocked_ioctl(struct file *file, unsigned int cmd, unsigned long arg)
{
	int retval;

	switch (cmd)
    {
    char tmpBuf[80];
    case IOCTL_VERSION:
        retval = sprintf(tmpBuf, DRIVER_DESC " " DRIVER_VERSION ", Linux %d.%d.%d",
            (LINUX_VERSION_CODE >> 16) & 0xFF,
            (LINUX_VERSION_CODE >> 8) & 0xFF,
            LINUX_VERSION_CODE & 0xFF ) + 1;
        if (copy_to_user ((char *)arg, tmpBuf, retval))
            retval = -EFAULT;
        break;
	default:
		retval = -EINVAL;
	}
	return retval;
}

